package ap.proof.goal;

import ap.terfor.{Formula, Conjunction, ArithConj, Quantifier}
import ap.util.Debug

object ComplexFormulaTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK

  
}

class ComplexFormulaTask(formula : Formula) extends Task {
  
  val priority : Int = -20000

  def apply(goal : Goal, taskFinishedAction : TaskFinishedAction) : ProofTree = {
    print("c")
    if (formula.isTrue) {
      // then the goal can be closed immediately
      goal.taskFinished(Conjunction.FALSE, taskFinishedAction)
    } else if (formula.isFalse) {
      // then the formula can be removed
      goal.taskFinished(taskFinishedAction)
    } else {
      // the complex formula has to be examined and possibly split into
      // multiple conjunctions (but without splitting the goal, this is done
      // by <code>ConjunctionTask</code>)
  
      val order = goal.order
      
      def extendedFacts(newFacts : Formula) : Conjunction =
        Conjunction.conj(Array(goal.facts, newFacts), order)
      
      def addLiteral(lit : Formula) : ProofTree =
        goal.taskFinished(extendedFacts(Conjunction.negate(lit, order)),
                          Array(FactsNormalisationTask),
                          taskFinishedAction)        

      def addTaskFor(conj : Conjunction) : ProofTree =
        goal.taskFinished(Array(createTaskFor(conj, goal.age)),
                          taskFinishedAction)

      formula match {
        
      case ac : ArithConj =>
      	if (ac.isLiteral)
          addLiteral(ac)
      	else
          addTaskFor(Conjunction.conj(ac, order))

      case conj : Conjunction =>
        if (conj.isLiteral) {
          addLiteral(conj)
        } else if (conj.isNegatedConjunction) {
          // apply alpha-rules
          
          val disj = conj.negatedConjs(0)
          //////////////////////////////////////////////////////////////////////
          // possibly existing quantifiers should have been pulled out and
          // should not occur at this point
          Debug.assertInt(ComplexFormulaTask.AC, disj.quans.isEmpty)
          //////////////////////////////////////////////////////////////////////
          
          // the literals of the negated conjunction can be added as facts to
          // the goal
          val updatedFacts = extendedFacts(disj.arithConj)
          
          // the nested conjunctions have to be handled recursively
          val newTasks = for (f <- disj.negatedConjs) yield new ComplexFormulaTask(f)
          
	  goal.taskFinished(updatedFacts,
                            newTasks ++ Array(FactsNormalisationTask),
                            taskFinishedAction)
        } else {
          addTaskFor(conj)
        }
      }
    }
  }
  
  private def createTaskFor(conj : Conjunction, age : Int) =
    if (conj.quans.isEmpty) {
      new ConjunctionTask(conj, age)
    } else conj.quans.last match {
      case Quantifier.ALL => new AllQuantifierTask(conj, age)
      case Quantifier.EX => new ExQuantifierTask(conj, age)
    }
}
