package ap.proof.goal;

import ap.basetypes.IdealInt
import ap.terfor.{Formula, Conjunction, Quantifier, ConstantTerm, VariableTerm}
import ap.terfor.substitutions.VariableSubst
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.linearcombination.LinearCombination
import ap.util.{Debug, PlainRange}

import scala.collection.mutable.ArrayBuffer

object ExQuantifierTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
}

class ExQuantifierTask(conj : Conjunction, age : Int)
      extends QuantifierTask(conj, age) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(ExQuantifierTask.AC, conj.quans.last == Quantifier.EX)
  //////////////////////////////////////////////////////////////////////////////
    
  protected val constantNameBase : String = "ex_"
    
  protected def newEliminatedConstants(goal : Goal,
                                       newConstants : Iterable[ConstantTerm])
                                               : Set[ConstantTerm] = Set.empty
  print("e")
  def apply(goal : Goal, taskFinishedAction : TaskFinishedAction) : ProofTree = {
    if (conj.isDivisibility)
      caseDistinction(goal, taskFinishedAction)
    else
      instantiateWithConstants(goal, taskFinishedAction)    
  }
  
  /**
   * Transform a formula <code>EX (n*_0 + t = 0)</code> in the succedent into
   * a conjunction
   * <code>ALL (n*_0 + t + 1 != 0) /\ ALL (n*_0 + t + 2 != 0) /\ ...</code>
   * in the succedent
   */
  private def caseDistinction(goal : Goal,
                              taskFinishedAction : TaskFinishedAction) : ProofTree = {
    val lc = conj.arithConj.positiveEqs(0)
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertInt(ExQuantifierTask.AC,
                    !lc.isEmpty && lc.leadingTerm == VariableTerm(0))
    ////////////////////////////////////////////////////////////////////////////
    
    val coeff : IdealInt = lc.leadingCoeff
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertInt(ExQuantifierTask.AC, coeff.signum > 0)
    ////////////////////////////////////////////////////////////////////////////
    
    val order = goal.order
    
    val cases = new ArrayBuffer[Formula]
    
    var offset : IdealInt = IdealInt.ONE
    while (offset < coeff) {
      val lcWithOffset = LinearCombination.sum(Array((IdealInt.ONE, lc),
                                                     (offset, LinearCombination.ONE)),
                                               order)
      cases += NegEquationConj(lcWithOffset, order)
      offset = offset + IdealInt.ONE
    }
    
    val caseDistinction = Conjunction.quantify(Array(Quantifier.ALL),
                                               Conjunction.conj(cases, order),
                                               order)
    
    goal.taskFinished(Array(new ComplexFormulaTask(caseDistinction)),
                      taskFinishedAction)
  }
}
