package ap.terfor.substitutions;

import ap.basetypes.IdealInt
import ap.util.Debug

object PseudoConstantSubst {
  
  private val AC = Debug.AC_SUBSTITUTIONS
  
}

/**
 * Replace a constant with an arbitrary term
 */
class PseudoConstantSubst(coeff : IdealInt,
                          constant : ConstantTerm,
                          replacement : Term,
                          protected [substitutions] val order : TermOrder)
      extends PseudoDivSubstitution with Sorted[PseudoConstantSubst] {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(PseudoConstantSubst.AC,
                   (order isSortingOf replacement) && !coeff.isZero)
  //////////////////////////////////////////////////////////////////////////////     

  protected[substitutions] def passQuantifiers(num : Int) : Substitution =
    new PseudoConstantSubst(coeff,
                            constant,
                            // TODO: could be optimised by caching
                            new VariableShiftSubst(0, num, order) (replacement),
                            order)
    
  protected[substitutions] def applyToVariable(v : VariableTerm) : (IdealInt, Term) =
    (IdealInt.ONE, v)

  protected[substitutions] def applyToConstant(c : ConstantTerm) : (IdealInt, Term) =
    if (c == constant) (coeff, replacement) else (IdealInt.ONE, c)

  protected[substitutions] def isIdentityOn(t : TerFor) : Boolean =
    !(t.constants contains constant)

  //////////////////////////////////////////////////////////////////////////////

  def sortBy(newOrder : TermOrder) : PseudoConstantSubst =
    new PseudoConstantSubst(coeff, constant, newOrder sort replacement, newOrder)

  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = "[" + constant + " |-> " + replacement + "]"

}
